/**
 * 依赖cordova-plugin-file和cordova-plugin-file-transfer两个插件
 * 
 * App热更新工具，将文件从可读路径下转移到缓存路径，通过判别updater.json
 * 来判断是否需要更新。如果确实需要更新则将需要更新的文件下载到临时目录下，
 * 删除缓存路径下的旧文件并使用新文件替换之。执行文件io操作时会让
 * START_IOCHANGE为true，如果有新的io操作时START_IOCHANGE仍然为true，
 * 则会完全删除缓存路径下的文件之后再进行操作。
 */

UpdateState.STATE_UNKNOW = 0;// 未知错误
UpdateState.STATE_DEBUGMODE = 1;// 未知错误
UpdateState.STATE_NORESOURCES = 2;// 找不到原始资源
UpdateState.STATE_NOSWAP = 3;// 没有储存空间或文件系统不存在
UpdateState.STATE_FINISHLOCALCHECK = 4;// 本地检查结束
UpdateState.STATE_CONNECTERROR = 5;// 服务器响应异常
UpdateState.STATE_SERVERERROR = 6;// 服务器响应异常
UpdateState.STATE_RESPONSEERROR = 7;// 报文异常
UpdateState.STATE_FINISH = 8;// 检查更新成功，不需要更新
UpdateState.STATE_UPDATE = 9;// 需要更新
UpdateState.STATE_ONUPDATE = 10;// 正在更新
UpdateState.STATE_FINISHDOWNLOAD = 11;// 下载更新文件结束
UpdateState.STATE_UPDATEFINISH = 12;// 更新文件成功
UpdateState.STATE_UPDATEREFUSE = 13;// 不支持更新到最新版本
UpdateState.STATE_UPDATENOALARM = 14;// 用户不希望升级至该版本

var htmlStorage = window.localStorage;
var DEBUG = false;
document.addEventListener("deviceready", onDeviceReady, false);

// 获取app更新类型
function getUpdateType() {
	return "occupant";
}

function onDeviceReady() {
//	alert('dsdasdsad ' + cordova.file);
////	var url = 'http://192.168.1.178:8080/Test/';
//	var url = 'http://192.168.250.231:38438/v1/static/www/';
////	var url = 'http://192.168.250.195:8080/Test/';
}

function UpdateState() {
	this.state = UpdateState.STATE_UNKNOW;
}

// 设置应用程序类型和版本号，程序类型android和iOS，版本号0.0.0格式
function setAppVersion(type, version) {
	if (type != 'android' && type != 'iOS') {
		console.log("Unlegal input type:" + type);
		return;
	}
	if (version && version.split('.').length != 3) {
		console.log("Unlegal input version:" + version);
		return;
	}
	htmlStorage.setItem("APP_TYPE", type);
	htmlStorage.setItem("APP_VERSION", version);
}

// 获取应用程序类型
function getAppType() {
	return htmlStorage.getItem("APP_TYPE");
}

// 获取应用程序类型
function getAppVersion() {
	return htmlStorage.getItem("APP_VERSION");
}

// app版本比较,，true需要更新
function versionCompare(webVersion) {
	var localVersion = getAppVersion();
	if (!localVersion) {
		return false;
	}
	localVersion = localVersion.split('.');
	webVersion = webVersion.split('.');
	if (localVersion[0] < webVersion[0]) {
		return true;
	} else if (localVersion[1] < webVersion[1]) {
		return true;
	} else if (localVersion[2] < webVersion[2]) {
		return true;
	}
	return false;
}

function setWWWVersion(version) {
	htmlStorage.setItem("WWW_VERSION", version);
}

var BASE_VERSION = 0;// 固化于App包中的原始版本，和updater.json中的version对应
function getWWWVersion() {
	var version =  htmlStorage.getItem("WWW_VERSION");
	if (version) {
		return version;
	}
	return BASE_VERSION;
}

function setLocalWWWConfig(config) {
	htmlStorage.setItem("WWW_CONFIG", JSON.stringify(config));
}

//固化于App包中的原始配置，相当于updater.json
var BASE_CONFIG = {
	"version" : 0,
	"app-version-android" : "0.0.0",
	"app-version-iOS" : "0.0.0",
	"update-link-android" : "",
	"update-link-iOS" : "",
	"inherit" : false,
	"update-now" : true,
	"files" : [],
	"start" : "index.html",
	"descriptions" : []
};
function getLocalWWWConfig() {
	var config = htmlStorage.getItem("WWW_CONFIG");
	if (!config) {
		return BASE_CONFIG;
	}
	try {
		config = JSON.parse(config);
	} catch (err) {
		console.warn(dump(err));
		return BASE_CONFIG;
	}
	return config;
}

function getLastVersionNoAlarm() {
	return window.localStorage.getItem("NEWVERSION_NOALARM");
}

function setNewVersionNoAlarm(version) {
	window.localStorage.setItem("NEWVERSION_NOALARM", version);
}

// 入口，检查是否需要更新
function checkVersion(ip, isCheckServer, stateCallback) {
	if (DEBUG) {
		stateCallback && stateCallback({ state: UpdateState.STATE_DEBUGMODE }, 0);
		return;
	}
	console.warn(cordova.file.applicationDirectory);
	window.resolveLocalFileSystemURL(cordova.file.dataDirectory , function(fileEntry) {
		checkLocalUpdater(fileEntry, false, function(localConfig) {
			setLocalWWWConfig(localConfig);
			setWWWVersion(localConfig.version);
			stateCallback && stateCallback({
				state: UpdateState.STATE_FINISHLOCALCHECK, 
				path: fileEntry.toURL() + 'www/' + localConfig.start
			}, 0);
			if (isCheckServer) {
				checkNewVersion(fileEntry, ip, localConfig, stateCallback);
			}
		}, function(err) {
			stateCallback && stateCallback({ state: UpdateState.STATE_NORESOURCES }, 0);
			console.warn(dump(err));
		}, 0);
	}, function(err) {
		stateCallback && stateCallback({ state: UpdateState.STATE_NOSWAP }, 0);
		console.warn(dump(err));
	});
}

// 检查本地www文件的版本
function checkLocalUpdater(wwwSwapDirectory, forceUpdate, successCallback, failCallback, n) {
	if (n > 3) {
		console.log("checkLocalUpdater exit after several failed attempts.");
		failCallback && failCallback({error: 'unknow', code: 110});
		return;
	}
	
	var targetPath = cordova.file.applicationDirectory + '/www';
	wwwSwapDirectory.getFile("www/updater.json", { create: false }, function (fileEntry) {
		if(htmlStorage.getItem("START_IOCHANGE") == 'true' || forceUpdate) {
			console.log("START_IOCHANGE is true get error, redo copyWWWToSwapDirectory.");
			removeRecursively(cordova.file.dataDirectory  + '/www', function(err) {
				if ((err) && (err != null)) {
					console.warn(dump(err));
	    			failCallback && failCallback(err);
				} else {
					copyWWWToSwapDirectory(targetPath, wwwSwapDirectory, function(err) {
			    		if ((err) && (err != null)) {
		            		console.warn(dump(err));
			    			failCallback && failCallback(err);
			    		} else {
			    			checkLocalUpdater(wwwSwapDirectory, false, successCallback, failCallback, n + 1);
			    		}
			    	});
				}
			});
		} else {
			checkLocalVersion(cordova.file.dataDirectory  + '/www/updater.json', function(swapConfig) {
				checkLocalVersion(cordova.file.applicationDirectory + '/www/updater.json', function(wwwConfig) {
					if (swapConfig.version < wwwConfig.version) {
						checkLocalUpdater(wwwSwapDirectory, true, successCallback, failCallback, n + 1);
					} else {
						successCallback && successCallback(swapConfig);
					}
				}, failCallback);
			}, failCallback);
		}
    }, function(err) {
    	console.log("Can not find " + wwwSwapDirectory.toURL() + "www/updater.json, try to call copyWWWToSwapDirectory and rebuild it.");
    	copyWWWToSwapDirectory(targetPath, wwwSwapDirectory, function(err) {
    		if ((err) && (err != null)) {
    			failCallback && failCallback(err);
    		} else {
    			checkLocalUpdater(wwwSwapDirectory, false, successCallback, failCallback, n + 1);
    		}
    	});
    });
}

// 删除指定目录及文件
function removeRecursively(targetPath, callback) {
	console.log('Remove recursively directory ' + targetPath);
	window.resolveLocalFileSystemURL(targetPath, function(fileEntry) {
		fileEntry.removeRecursively(function() {
			callback && callback();
		}, function(err) {
			console.log('Can not removeRecursively path' + targetPath + '. code ' + err.code);
			callback && callback(err);
		});
	}, function(err) {
		console.log('Can not found the path' + targetPath + '. code ' + err.code);
		callback && callback(err);
	});
}

// 将只读文件夹内的www文件拷贝到可写的位置
function copyWWWToSwapDirectory(targetPath, parent, callback) {
	console.log('Copy www file from' + targetPath);
	window.resolveLocalFileSystemURL(targetPath, function(fileEntry) {
		htmlStorage.setItem("START_IOCHANGE", true);
		fileEntry.copyTo(parent, "www", function(targetDirectory) {
			htmlStorage.setItem("START_IOCHANGE", false);
			callback && callback();
		}, function(err) {
			callback && callback(err);
		});
	}, function(err) {
		console.log('Can not found the original www resources when copyWWWToSwapDirectory, do nothing and return. code ' + err.code);
		callback && callback(err);
	});
}

// 拷贝文件列表
function copyFileList(targetPath, parent, files, index, callback) {
	if (!files || files.length <= index) {
		callback && callback();
		return;
	}
	
	var file = files[index];
	console.log('Copy file' + targetPath + file + ' to ' + parent.toURL() + "www/" + file);
	window.resolveLocalFileSystemURL(targetPath + file, function(fileEntry) {
		fileEntry.copyTo(parent, "www/" + file, function(targetDirectory) {
			copyFileList(targetPath, parent, files, index + 1, callback);
		}, function(err) {
			callback && callback(err);
		});
	}, function(err) {
		console.log('Can not copy file ' + targetPath + file + '. code ' + err.code);
		callback && callback(err);
	});
}

// 移除指定文件夹内的部分文件
function removeWWWFileList(parent, files, index, callback) {
	if (!files || files.length <= index) {
		callback && callback();
		return;
	}
	
	var file = files[index];
	console.log('Start remove old swap file ' + parent.toURL() + "www/" + file);
	parent.getFile("www/" + file, { create: false }, function (subDirEntry) {
		subDirEntry.remove(function() {
			removeWWWFileList(parent, files, index + 1, callback);
		}, function(err) {
			console.log('removeRecursively ' + fileEntry.toURL() + "www" + ' fail. code ' + err.code);
			callback && callback(err);
		});
    }, function(err) {
		console.log('Can not found the ' + parent.toURL() + "www/" + file + ' when removeSwapDirectoryRecursively. code ' + err.code);
		removeWWWFileList(parent, files, index + 1, callback);
	});
}

// 检查本地swap文件夹内的json文件版本
function checkLocalVersion(url, successCallback, failCallback) {
	window.resolveLocalFileSystemURL(url, function(fileEntry) {
		fileEntry.file(function(file) {
			var reader = new FileReader();
            reader.onloadend = function (evt) {
                try {
                	successCallback && successCallback(JSON.parse(evt.target.result));
                } catch (er) {
                	failCallback && failCallback({error: 'json parseFail', code: 111});
                }
            };
            reader.readAsText(file);
	    }, function(err) {
			console.log('Can not get file when checkLocalVersion, do nothing and return. code ' + err.code);
			failCallback && failCallback(err);
	    });
	}, function(err) {
		console.log('Can not found the original www resources when checkLocalVersion, do nothing and return. code ' + err.code);
		failCallback && failCallback(err);
	});
}

// 按顺序下载数组中的文件
function downloadInOrder(fileEntry, updateUrl, files, index, callback, progressCallback) {
	if (!files) {
		progressCallback && progressCallback(1);
		return;
	}
	progressCallback && progressCallback(files.length != 0 ? index / files.length : 1);
	if (files.length <= index) {
		return;
	}

	var file = files[index];
	console.log('Ready to download file, the temporary path is ' + fileEntry.toURL() + '/temp/' + file);
	download(updateUrl + file, fileEntry.toURL() + '/temp/' + file, function(err) {
		if ((err) && (err != null)) {
			callback && callback(err);
		} else {
			downloadInOrder(fileEntry, updateUrl, files, index + 1, callback, progressCallback);
		}
    });
}

// 开始下载
function download(uri, fileURL, callback) {

    var fileTransfer = new FileTransfer();

    fileTransfer.download(
        uri,
        fileURL,
        function (entry) {
            console.log("Successful download...");
            console.log("download complete: " + entry.toURL());
            callback && callback();
        },
        function (error) {
        	console.log("download error source " + error.source);
            console.log("download error target " + error.target);
            console.log("upload error code" + error.code);
            callback && callback(error);
        },
        null, // or, pass false
        {
            //headers: {
            //    "Authorization": "Basic dGVzdHVzZXJuYW1lOnRlc3RwYXNzd29yZA=="
            //}
        }
    );
}

// 向服务器拉取最新版本文件
function checkNewVersion(fileEntry, ip, localConfig, stateCallback) {
	var updateUrl = ip + "/v1/update?v=" + getWWWVersion() + "&type=" + getUpdateType();
	try {
		var request = new XMLHttpRequest();
		console.log('start ajax connect to ' + updateUrl);
	    request.open('GET', updateUrl);
	    request.onreadystatechange = function (e) {
	    	console.log('readyState ' + this.readyState + ' status ' + this.status + ' responseText ' + this.responseText);
	        if (this.readyState == 4) {
	            if (this.status == 200) {
	            	// 成功获取更新信息报文
	                try {
	                    var config = JSON.parse(this.responseText).result;
	                    console.log("local version " + localConfig.version + ", web version " + config.version);
                    	if (getLastVersionNoAlarm() == config.version && mainView.activePage.name != 'checkUpdate') {// 用户不希望升级至该版本
	                    	stateCallback && stateCallback({
	            				state: UpdateState.STATE_UPDATENOALARM, 
	            				link: type && config['update-link-' + type], 
	            				descriptions: config.descriptions, 
	            				version: config.version
	            			}, 0);
                    	} else if (localConfig.version < config.version) {
	                    	var type = getAppType();
	                        if (!type || versionCompare(config['app-version-' + type])) {// 该版本需要更新的软件版本支持，无法更新
	                        	stateCallback && stateCallback({
	                				state: UpdateState.STATE_UPDATEREFUSE, 
	                				link: type && config['update-link-' + type], 
	                				descriptions: config.descriptions, 
	                				version: config.version
	                			}, 0);
	                        } else if (!config.inherit) {// 该版本不支持低版本向上更新
	                        	stateCallback && stateCallback({
	                				state: UpdateState.STATE_UPDATEREFUSE, 
	                				link: type && config['update-link-' + type], 
	                				descriptions: config.descriptions, 
	                				version: config.version
	                			}, 0);
	                        } else {// 正常的更新步骤
	                        	// 如果不存在就把updater.json加进列表，updater.json是更新时必更的
	                        	var hasUpdateJson = true;
	                        	for (var i = 0; i < config.files.length; i++) {
	                                if (config.files[i] == "updater.json") {
	                                	hasUpdateJson = false;
	                                	break;
	                                }  
	                            }
	                        	if (hasUpdateJson) {
	                        		config.files.push("updater.json");
	                        	}
	                        	// 本地的是旧版本
	                			stateCallback && stateCallback({ state: UpdateState.STATE_UPDATE }, 0);
	                        	downloadInOrder(fileEntry, ip + '/v1/static/' + getUpdateType() + '/', config.files, 0, function(err) {
	                        		console.warn(dump(err));
	                        		stateCallback && stateCallback({ state: UpdateState.STATE_SERVERERROR }, 0);
	                        	}, function(progress) {
	                        		// 进度回调
	                        		if (progress == 1) {
	                        			stateCallback && stateCallback({
	                        				state: UpdateState.STATE_FINISHDOWNLOAD, 
	                        				"update-now": config['update-now'], 
	                        				handler: createUpdateSpecifiedFilesClosure(fileEntry, config, stateCallback), 
	    	                				descriptions: config.descriptions, 
	    	                				version: config.version
	                        			}, progress);
//	                        			updateSpecifiedFiles(fileEntry, config, stateCallback);
	                        		} else {
	                        			stateCallback && stateCallback({ state: UpdateState.STATE_ONUPDATE }, progress);
	                        		}
	                        	});
	                        }
	                    } else {
	                    	stateCallback && stateCallback({ state: UpdateState.STATE_FINISH }, 0);
	                    }
	                } catch (e) {
	                	dump(e);
	    				if (this.responseText) {
	    					stateCallback && stateCallback({ state: UpdateState.STATE_RESPONSEERROR }, 0);
	    	    			console.warn("parse json error, response: " + this.responseText);
	    				} else {
	    					stateCallback && stateCallback({ state: UpdateState.STATE_SERVERERROR }, 0);
	    	    			console.warn("connect error");
	    				}
	                }
	            } else {
					stateCallback && stateCallback({ state: UpdateState.STATE_SERVERERROR }, 0);
	    			console.warn("http get readyState " + this.readyState + " status " + this.status);
	            }
	        } else {
				stateCallback && stateCallback({ state: UpdateState.STATE_SERVERERROR }, 0);
    			console.warn("http get readyState " + this.readyState + " status " + this.status);
	        }
	    };
	    request.send(null);
    } catch (err) {
    	console.warn("if you catch connect refused, set you ip in html 'Content-Security-Policy'.");
    	stateCallback && stateCallback({ state: UpdateState.STATE_CONNECTERROR }, 0);
    }
}

// 创建闭包，包含函数updateSpecifiedFiles
function createUpdateSpecifiedFilesClosure(fileEntry, config, stateCallback) {
	function closure() {
		showPreLoader && showPreLoader();// 打开加载progress
		updateSpecifiedFiles(fileEntry, config, stateCallback)
	}
	return closure;
}

// 移除旧版本上的文件并将新文件移至工作目录
function updateSpecifiedFiles(fileEntry, config, stateCallback) {
	var targetPath = cordova.file.dataDirectory  + '/temp';
	htmlStorage.setItem("START_IOCHANGE", true);
	removeWWWFileList(fileEntry, config.files, 0, function(err) {
		if ((err) && (err != null)) {
			hidePreLoader && hidePreLoader();// 关闭加载progress
			stateCallback && stateCallback({ state: UpdateState.STATE_NOSWAP }, 0);
		} else {
			var targetPath = cordova.file.dataDirectory  + '/temp/';
			copyFileList(targetPath, fileEntry, config.files, 0, function(err) {
				hidePreLoader && hidePreLoader();// 关闭加载progress
	    		if ((err) && (err != null)) {
            		console.warn(dump(err));
	    			stateCallback && stateCallback({ state: UpdateState.STATE_UNKNOW }, 0);
	    		} else {
	    			stateCallback && stateCallback({
	    				state: UpdateState.STATE_UPDATEFINISH, 
	    				path: config.start
	    			}, 0);
					htmlStorage.setItem("START_IOCHANGE", false);
	    		}
	    	});
		}
	});
}

// 打印结构
function dump(myObject) {
	for(var property in myObject) {
		console.warn("" + property + ": " + myObject[property]);
	}
}